--[[
Copyright 2012 João Cardoso
PetTracker is distributed under the terms of the GNU General Public License (Version 3).
As a special exception, the copyright holders of this addon do not give permission to
redistribute and/or modify it.

This addon is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the addon. If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.

This file is part of PetTracker.
--]]

local ADDON, Addon = ...
local Journal = Addon.Journal
local BlipParent, MapFrame = WorldMapButton, WorldMapDetailFrame

local Map = Addon:NewModule('WorldMap', CreateFrame('Button', ADDON..'MapButton', BlipParent, 'PetTrackerMapButton'))
local Tooltip = Addon.MapTip()


--[[ Startup ]]--

function Map:Startup()
	self:RegisterEvent('MINIMAP_UPDATE_TRACKING')
	self.Flyout:SetScript('OnClick', self.ShowDrop)
	self:SetScript('OnClick', self.Toggle)

	self:SetScript('OnEvent', self.UpdateBlips)
	self:SetScript('OnUpdate', self.UpdateTip)
	self:SetScript('OnHide', self.HideTip)
	
	self:SetActive(self:IsActive())
	self:SetPoint('TOPRIGHT', -6, -10)
	self:SetScale(1.1)
	self.blips = {}
	
	hooksecurefunc('WorldMapFrame_Update', function()
		self:UpdateBlips()
	end)
end


--[[ Blips ]]--

function Map:UpdateBlips()
	self:Reset()

	if self:IsActive() then
		self:ShowBlips()
	end
end

function Map:ShowBlips()
	local width, height = MapFrame:GetSize()
	local species = Journal:GetCurrentSpecies()
	
	for specie, spots in pairs(species) do
		local best = Journal:GetBestOwned(specie)
		
		if best <= self:GetFilter() then
			local icon = Journal:GetTypeIcon(specie)

			for x, y in gmatch(spots, '(%w+):(%w+)') do
				local x, y = self:ParseCoords(x, y)
				local blip = Addon.SpecieBlip(BlipParent)
				blip:SetPoint('CENTER', MapFrame, 'TOPLEFT', x * width, -y * height)
				blip.icon:SetTexture(icon)
				blip.x, blip.y = x, y
				blip.specie = specie
				blip:Show()

				tinsert(self.blips, blip)
			end
		end
	end
end

function Map:ParseCoords(x, y)
	return tonumber(x, 16) / 1000, tonumber(y, 16) / 1000
end

function Map:Reset()
	for _, blip in ipairs(self.blips) do
		blip:Release()
	end

	wipe(self.blips)
end


--[[ Tooltip ]]--

function Map:UpdateTip()
	Tooltip:Anchor(BlipParent, 'ANCHOR_CURSOR')

	for i, blip in ipairs(self.blips) do
		if blip:IsMouseOver() then
			local title, text = blip:GetTooltip()
			
			Tooltip:AddHeader(title)
			Tooltip:AddLine(text, 1,1,1, true)
		end
	end

	Tooltip:Display()
end

function Map:HideTip()
	Tooltip:Hide()
end


--[[ Toggle ]]--

function Map:Toggle()
	self:SetActive(not self:IsActive())
	self:UpdateBlips()
end

function Map:SetActive(active)
	local disabled = not active
	PetTracker_HideSpots = disabled
	
	self.Icon:SetDesaturated(disabled)
	self.Flyout:GetNormalTexture():SetDesaturated(disabled)
	self.Flyout:GetHighlightTexture():SetDesaturated(disabled)
end

function Map:IsActive()
	return not PetTracker_HideSpots
end


--[[ Filter ]]--

function Map:ShowDrop()
	local info = {
		{
			text = 'Pets',
			notCheckable = 1,
			isTitle = 1
    	}
	}
	
	local function addLine(name, quality)
		tinsert(info, {
			text = name,
			func = function()
				PetTracker_QualityFilter = quality
				Map:UpdateBlips()
			end
    	})
	end
	
	addLine('All', Addon.MaxQuality)
	addLine('Not Maximized', Addon.MaxQuality - 1)
	addLine('Missing', 0)

	EasyMenu(info, Addon.Blip.Drop, self, 0, 0)
end

function Map:GetFilter()
	return PetTracker_QualityFilter or Addon.MaxQuality
end